//
// main.c
// The generic launcher for Shoes.
//
#include <stdio.h>
#include "shoes/app.h"
#include "shoes/ruby.h"
#include "shoes/config.h"
#include "shoes/world.h"
#include "shoes/internal.h"

#ifdef __APPLE__
#include <crt_externs.h>
#endif

#ifdef SHOES_WIN32
int WINAPI
WinMain(HINSTANCE inst, HINSTANCE inst2, LPSTR arg, int style)
#else
int
main(argc, argv)
  int argc;
  char *argv[];
#endif
{
  shoes_code code;
  char *path = NULL;
#ifdef __APPLE__
  char **env = *_NSGetEnviron();
#endif
#ifdef SHOES_WIN32
  int argc;
  char **argv;
  argc = shoes_win32_cmdvector(GetCommandLine(), &argv);
#endif


#ifdef SHOES_WIN32
  path = SHOE_ALLOC_N(char, SHOES_BUFSIZE);
  GetModuleFileName(NULL, (LPSTR)path, SHOES_BUFSIZE);
  code = shoes_init(inst, style);
#else
  path = argv[0];
  if (argc > 1 && strcmp(argv[1], "--ruby") == 0)
  {
    char bootup[SHOES_BUFSIZE];
    int len = shoes_snprintf(bootup,
      SHOES_BUFSIZE,
      "begin;"
        "DIR = File.expand_path(File.dirname(%%q<%s>));"
        "$:.replace([DIR+'/ruby/lib/'+PLATFORM, DIR+'/ruby/lib', DIR+'/lib', '.']);"
        "require 'shoes/cache';"
        "DIR;"
      "rescue Object => e;"
        "puts(e.message);"
      "end",
      path);

    if (len < 0 || len >= SHOES_BUFSIZE)
      return 0;

    argc--;
    argv[1] = argv[0];
    argv = &argv[1];
    {
      RUBY_INIT_STACK
      ruby_init();
      rb_eval_string(bootup);
      ruby_options(argc, argv);
      ruby_run();
    }
    return 0;
  }

  code = shoes_init();
#endif
  if (code != SHOES_OK)
    goto done;

  shoes_set_argv(argc - 1, &argv[1]);
  code = shoes_start(path, "/");
  if (code != SHOES_OK)
    goto done;

done:
#ifdef SHOES_WIN32
  if (path != NULL)
    SHOE_FREE(path);
#endif
  shoes_final();
  return 0;
}
